import numpy as np
from . import VBHelperFunctions
from . import BranchingTree as bt
from . import branch_kernParamGPflow as bk
from . import assigngp_denseSparse
from . import assigngp_dense
import gpflow
import traceback
import sys

def FitModel(bConsider, GPt, GPy, globalBranching, priorConfidence=0.80,
             M=10, likvar=1., kerlen=2., kervar=5., fDebug=False, maxiter=100,
             fPredict=True, fixHyperparameters=False):
    """
    Fit BGP model
    :param bConsider: list of candidate branching points
    :param GPt: pseudotime
    :param GPy: gene expression. Should be 0 mean for best performance.
    :param globalBranching: cell labels
    :param priorConfidence: prior confidence on cell labels
    :param M: number of inducing points
    :param likvar: initial value for Gaussian noise variance
    :param kerlen: initial value for kernel length scale
    :param kervar: initial value for kernel variance
    :param fDebug: Print debugging information
    :param maxiter: maximum number of iterations for optimisation
    :param fPredict: compute predictive mean and variance
    :param fixHyperparameters: should kernel hyperparameters be kept fixed or optimised?
    :return: dictionary of log likelihood, GPflow model, Phi matrix, predictive set of points,
    mean and variance, hyperparameter values, posterior on branching time
    """
    assert isinstance(bConsider, list), 'Candidate B must be list'
    assert GPt.ndim == 1
    assert GPy.ndim == 2
    assert GPt.size == GPy.size, 'pseudotime and gene expression data must be the same size'
    assert globalBranching.size == GPy.size, 'state space must be same size as number of cells'
    assert M >= 0, 'at least 0 or more inducing points should be given'
    phiInitial, phiPrior = GetInitialConditionsAndPrior(globalBranching, priorConfidence, infPriorPhi=True)
    # it = np.argsort(GPt)
    # print('GPt and prior', np.hstack([GPt[it][:,None], phiPrior[it,:], phiInitial[it,:]]))
    # print(phiPrior[:5, :])
    XExpanded, indices, _ = VBHelperFunctions.GetFunctionIndexListGeneral(GPt)
    ptb = np.min([np.min(GPt[globalBranching == 2]), np.min(GPt[globalBranching == 3])])
    tree = bt.BinaryBranchingTree(0, 1, fDebug=False)
    tree.add(None, 1, np.ones((1, 1)) * ptb)  # B can be anything here
    (fm, _) = tree.GetFunctionBranchTensor()
    with gpflow.defer_build():
        kb = bk.BranchKernelParam(gpflow.kernels.Matern32(1), fm, b=np.zeros((1, 1))) + gpflow.kernels.White(1)
        kb.kernels[1].variance = 1e-6  # controls the discontinuity magnitude, the gap at the branching point
        kb.kernels[1].variance.set_trainable(False)  # jitter for numerics
        if(M == 0):
            m = assigngp_dense.AssignGP(GPt, XExpanded, GPy, kb, indices,
                                                    np.ones((1, 1)) * ptb, phiInitial=phiInitial,
                                                    phiPrior=phiPrior)
        else:
            ZExpanded = np.ones((M, 2))
            ZExpanded[:, 0] = np.linspace(0, 1, M, endpoint=False)
            ZExpanded[:, 1] = np.array([i for j in range(M) for i in range(1, 4)])[:M]
            m = assigngp_denseSparse.AssignGPSparse(GPt, XExpanded, GPy, kb, indices,
                                                    np.ones((1, 1)) * ptb, ZExpanded, phiInitial=phiInitial, phiPrior=phiPrior)
        # Initialise hyperparameters
        m.likelihood.variance = likvar
        m.kern.kernels[0].kern.lengthscales = kerlen
        m.kern.kernels[0].kern.variance = kervar
        if(fixHyperparameters):
            print('Fixing hyperparameters')
            m.kern.kernels[0].kern.lengthscales.set_trainable(False)
            m.likelihood.variance.set_trainable(False)
            m.kern.kernels[0].kern.variance.set_trainable(False)
        else:
            if fDebug:
                print('Adding prior logistic on length scale to avoid numerical problems')
            m.kern.kernels[0].kern.lengthscales.prior = gpflow.priors.Gaussian(2, .1)
            m.kern.kernels[0].kern.variance.prior = gpflow.priors.Gaussian(3, 1)
            m.likelihood.variance.prior = gpflow.priors.Gaussian(0.1, .1)
        m.compile()

    # optimization
    ll = np.zeros(len(bConsider))
    Phi_l = list()
    ttestl_l, mul_l, varl_l = list(), list(), list()
    hyps = list()
    for ib, b in enumerate(bConsider):
        m.UpdateBranchingPoint(np.ones((1, 1)) * b, phiInitial)
        try:
            gpflow.train.ScipyOptimizer().minimize(m, maxiter=maxiter)
            # remember winning hyperparameter
            hyps.append({'likvar':  m.likelihood.variance.value, 'kerlen':  m.kern.kernels[0].kern.lengthscales.value,
                    'kervar': m.kern.kernels[0].kern.variance.value})
            ll[ib] = m.compute_log_likelihood()
        except:
            print('Failure', "Unexpected error:", sys.exc_info()[0])
            print('-' * 60)
            traceback.print_exc(file=sys.stdout)
            print('Exception caused by model')
            print(m)
            print('-' * 60)
            ll[0] = np.nan
            # return model so can inspect model
            return {'loglik': ll, 'model': m, 'Phi': np.nan,
                    'prediction': {'xtest': np.nan, 'mu': np.nan, 'var': np.nan},
                    'hyperparameters': np.nan, 'posteriorB': np.nan}
        # prediction
        Phi = m.GetPhi()
        Phi_l.append(Phi)
        if(fPredict):
            ttestl, mul, varl = VBHelperFunctions.predictBranchingModel(m)
            ttestl_l.append(ttestl), mul_l.append(mul), varl_l.append(varl)
        else:
            ttestl_l.append([]), mul_l.append([]), varl_l.append([])
    iw = np.argmax(ll)
    postB = GetPosteriorB(ll, bConsider)
    if fDebug:
        print('BGP Maximum at b=%.2f' % bConsider[iw], 'CI= [%.2f, %.2f]' %(postB['B_CI'][0], postB['B_CI'][1]))
    assert np.allclose(bConsider[iw], postB['Bmode']), '%s-%s' % str(postB['B_CI'], bConsider[iw])
    return {'loglik': ll, 'Phi': Phi_l[iw], # 'model': m,
            'prediction': {'xtest': ttestl_l[iw], 'mu': mul_l[iw], 'var': varl_l[iw]},
            'hyperparameters': hyps[iw], 'posteriorB': postB}



def GetPosteriorB(objUnsorted, BgridSearch, ciLimits=[0.01, 0.99]):
    '''
    Return posterior on B for each experiment, confidence interval index, map index
    '''
    # for each trueB calculate posterior over grid
    # ... in a numerically stable way
    assert objUnsorted.size == len(BgridSearch), 'size do not match %g-%g' % (objUnsorted.size, len(BgridSearch))
    gr = np.array(BgridSearch)
    isort = np.argsort(gr)
    gr = gr[isort]
    o = objUnsorted[isort].copy()  # sorted objective funtion
    imode = np.argmax(o)
    pn = np.exp(o - np.max(o))
    p = pn/pn.sum()
    assert np.any(~np.isnan(p)), 'Nans in p! %s' % str(p)
    assert np.any(~np.isinf(p)), 'Infinities in p! %s' % str(p)
    pb_cdf = np.cumsum(p)
    confInt = np.zeros(len(ciLimits), dtype=int)
    for pb_i, pb_c in enumerate(ciLimits):
        pb_idx = np.flatnonzero(pb_cdf <= pb_c)
        if(pb_idx.size == 0):
            confInt[pb_i] = 0
        else:
            confInt[pb_i] = np.max(pb_idx)
    # if((imode+5) > 0 and imode < (len(BgridSearch)-5)):  # for modes at end points conf interval checks do not hold
    #     assert confInt[0] <= (imode-1), 'Lower confidence point bigger than mode! (%s)-%g' % (str(confInt), imode)
    #     assert confInt[1] >= (imode+1), 'Upper confidence point bigger than mode! (%s)-%g' % (str(confInt), imode)
    assert np.all(confInt < o.size), confInt
    B_CI = gr[confInt]
    Bmode = gr[imode]
    # return confidence interval as well as mode, and indexes for each
    return {'B_CI': B_CI, 'Bmode': Bmode, 'idx_confInt': confInt, 'idx_mode': imode, 'BgridSearch_sort': gr, 'isort': isort}




def GetInitialConditionsAndPrior(globalBranching, v, infPriorPhi):
    # Setting initial phi
    np.random.seed(42)  # UNDONE remove TODO
    assert isinstance(v, float), 'v should be scalar is %s' % str(type(v))
    N = globalBranching.size
    phiInitial = np.ones((N, 2))*0.5  # don't know anything
    phiInitial[:, 0] = np.random.rand(N)
    phiInitial[:, 1] = 1-phiInitial[:, 0]
    phiPrior = np.ones_like(phiInitial) * 0.5  # don't know anything
    for i in range(N):
        iBranch = globalBranching[i]-2  # is 1,2,3-> -1, 0, 1
        if(iBranch == -1):
            # trunk - set all equal
            phiPrior[i, :] = 0.5
        else:
            if(infPriorPhi):
                phiPrior[i, :] = 1-v
                phiPrior[i, int(iBranch)] = v
            phiInitial[i, int(iBranch)] = 0.5 + (np.random.random() / 2.)  # number between [0.5, 1]
            phiInitial[i, int(iBranch) != np.array([0, 1])] = 1 - phiInitial[i, int(iBranch)]
    assert np.allclose(phiPrior.sum(1), 1), 'Phi Prior should be close to 1 but got %s' % str(phiPrior)
    assert np.allclose(phiInitial.sum(1), 1), 'Phi Initial should be close to 1 but got %s' % str(phiInitial)
    assert np.all(~np.isnan(phiInitial)), 'No nans please!'
    assert np.all(~np.isnan(phiPrior)), 'No nans please!'
    return phiInitial, phiPrior



# coding: utf-8
import gpflow
import numpy as np
import tensorflow as tf
from . import pZ_construction_singleBP

from gpflow.params import DataHolder
from gpflow.params import Parameter
from gpflow.mean_functions import Zero
from gpflow import settings
from gpflow.decors import params_as_tensors, autoflow
from gpflow.models.model import GPModel

class AssignGP(GPModel):
    """
    Gaussian Process regression, but where the index to which the data are
    assigned is unknown.

    let f be a vector of GP points (usually longer than the number of data)

        f ~ MVN(0, K)

    and let Z be an (unknown) binary matrix with a single 1 in each row. The
    likelihood is

       y ~ MVN( Z f, \sigma^2 I)

    That is, each element of y is a noisy realization of one (unknown) element
    of f. We use variational Bayes to infer the labels using a sparse prior
    over the Z matrix (i.e. we have narrowed down the choice of which function
    values each y is drawn from).

    """

    def __init__(self, t, XExpanded, Y, kern, indices, b, phiPrior=None, phiInitial=None, fDebug=False, KConst=None):
        GPModel.__init__(self, XExpanded, Y, kern,
                                      likelihood=gpflow.likelihoods.Gaussian(),
                                      mean_function=Zero())
        assert len(indices) == t.size, 'indices must be size N'
        assert len(t.shape) == 1, 'pseudotime should be 1D'
        self.N = t.shape[0]
        self.t = t.astype(settings.float_type) # could be DataHolder? advantages
        self.indices = indices
        self.logPhi = Parameter(np.random.randn(t.shape[0], t.shape[0] * 3))  # 1 branch point => 3 functions
        if(phiInitial is None):
            phiInitial = np.ones((self.N, 2))*0.5  # dont know anything
            phiInitial[:, 0] = np.random.rand(self.N)
            phiInitial[:, 1] = 1-phiInitial[:, 0]
        self.fDebug = fDebug
        # Used as p(Z) prior in KL term. This should add to 1 but will do so after UpdatePhPrior
        if(phiPrior is None):
            phiPrior = np.ones((self.N, 2)) * 0.5
        # Fix prior term - this is without trunk
        self.pZ = DataHolder(np.ones((t.shape[0], t.shape[0] * 3)))
        self.UpdateBranchingPoint(b, phiInitial, prior=phiPrior)
        self.KConst = KConst
        if(not fDebug):
            assert KConst is None, 'KConst only for debugging'

    def UpdateBranchingPoint(self, b, phiInitial, prior=None):
        ''' Function to update branching point and optionally reset initial conditions for variational phi'''
        assert isinstance(self.pZ, DataHolder), 'Must have DataHolder'
        assert isinstance(b, np.ndarray)
        assert b.size == 1, 'Must have scalar branching point'
        self.b = b.astype(settings.float_type)  # remember branching value
        assert self.kern.kernels[0].name == 'BranchKernelParam'
        self.kern.kernels[0].Bv = b
        assert isinstance(self.kern.kernels[0].Bv, DataHolder)
        assert self.logPhi.trainable is True, 'Phi should not be constant when changing branching location'
        if prior is not None:
            self.eZ0 = pZ_construction_singleBP.expand_pZ0Zeros(prior)
        self.pZ = pZ_construction_singleBP.expand_pZ0PureNumpyZeros(self.eZ0, b, self.t)
        assert isinstance(self.pZ, DataHolder), 'Must have DataHolder'
        self.InitialiseVariationalPhi(phiInitial)

    def InitialiseVariationalPhi(self, phiInitialIn):
        ''' Set initial state for Phi using branching location to constrain.
        This code has to be consistent with pZ_construction.singleBP.make_matrix to where
        the equality is placed i.e. if x<=b trunk and if x>b branch or vice versa. We use the
         former convention.'''
        assert np.allclose(phiInitialIn.sum(1), 1), 'probs must sum to 1 %s' % str(phiInitialIn)
        assert self.b == self.kern.kernels[0].Bv.value, 'Need to call UpdateBranchingPoint'
        N = self.Y.value.shape[0]
        assert phiInitialIn.shape[0] == N
        assert phiInitialIn.shape[1] == 2  # run OMGP with K=2 trajectories
        phiInitialEx = np.zeros((N, 3 * N))
        # large neg number makes exact zeros, make smaller for added jitter
        phiInitial_invSoftmax = -9. * np.ones((N, 3 * N))
        eps = 1e-9
        iterC = 0
        for i, p in enumerate(self.t):
            if(p > self.b):  # before branching - it's the root
                phiInitialEx[i, iterC:iterC + 3] = np.hstack([eps, phiInitialIn[i, :] - eps])
            else:
                phiInitialEx[i, iterC:iterC + 3] = np.array([1 - 2 * eps, 0 + eps, 0 + eps])
            phiInitial_invSoftmax[i, iterC:iterC + 3] = np.log(phiInitialEx[i, iterC:iterC + 3])
            iterC += 3
        assert not np.any(np.isnan(phiInitialEx)), 'no nans please ' + str(np.nonzero(np.isnan(phiInitialEx)))
        assert not np.any(phiInitialEx < -eps), 'no negatives please ' + str(np.nonzero(np.isnan(phiInitialEx)))
        self.logPhi = phiInitial_invSoftmax

    def GetPhi(self):
        ''' Get Phi matrix, collapsed for each possible entry '''
        assert self.b == self.kern.kernels[0].Bv.value, 'Need to call UpdateBranchingPoint'
        phiExpanded = self.GetPhiExpanded()
        l = [phiExpanded[i, self.indices[i]] for i in range(len(self.indices))]
        phi = np.asarray(l)
        tolError = 1e-6
        assert np.all(phi.sum(1) <= 1+tolError)
        assert np.all(phi >= 0-tolError)
        assert np.all(phi <= 1+tolError)
        return phi

    @autoflow()
    @gpflow.params_as_tensors
    def GetPhiExpanded(self):
        ''' Shortcut function to get Phi matrix out.'''
        return tf.nn.softmax(self.logPhi)

    def objectiveFun(self):
        ''' Objective function to minimize - log likelihood -log prior.
        Unlike _objective, no gradient calculation is performed.'''
        return -self.compute_log_likelihood()-self.compute_log_prior()

    @params_as_tensors
    def _build_likelihood(self):
        print('assignegp_dense compiling model (build_likelihood)')
        N = tf.cast(tf.shape(self.Y)[0], dtype=settings.float_type)
        M = tf.shape(self.X)[0]
        D = tf.cast(tf.shape(self.Y)[1], dtype=settings.float_type)
        if(self.KConst is not None):
            K = tf.cast(self.KConst, settings.float_type)
        else:
            K = self.kern.K(self.X)
        Phi = tf.nn.softmax(self.logPhi)
        # try squashing Phi to avoid numerical errors
        Phi = (1 - 2e-6) * Phi + 1e-6
        sigma2 = self.likelihood.variance
        tau = 1. / self.likelihood.variance
        L = tf.cholesky(K) + tf.eye(M, dtype=settings.float_type) * settings.numerics.jitter_level
        W = tf.transpose(L) * tf.sqrt(tf.reduce_sum(Phi, 0)) / tf.sqrt(sigma2)
        P = tf.matmul(W, tf.transpose(W)) + tf.eye(M, dtype=settings.float_type)
        R = tf.cholesky(P)
        PhiY = tf.matmul(tf.transpose(Phi), self.Y)
        LPhiY = tf.matmul(tf.transpose(L), PhiY)
        if(self.fDebug):
            Phi = tf.Print(Phi, [tf.shape(P), P], message='P=', name='P', summarize=10)
            Phi = tf.Print(Phi, [tf.shape(LPhiY), LPhiY], message='LPhiY=', name='LPhiY', summarize=10)
            Phi = tf.Print(Phi, [tf.shape(K), K], message='K=', name='K', summarize=10)
            Phi = tf.Print(Phi, [tau], message='tau=', name='tau', summarize=10)
        c = tf.matrix_triangular_solve(R, LPhiY, lower=True) / sigma2
        # compute KL
        KL = self.build_KL(Phi)
        a1 = -0.5 * N * D * tf.log(2. * np.pi / tau)
        a2 = - 0.5 * D * tf.reduce_sum(tf.log(tf.square(tf.diag_part(R))))
        a3 = - 0.5 * tf.reduce_sum(tf.square(self.Y)) / sigma2
        a4 = + 0.5 * tf.reduce_sum(tf.square(c))
        a5 = - KL
        if(self.fDebug):
            a1 = tf.Print(a1, [a1], message='a1=')
            a2 = tf.Print(a2, [a2], message='a2=')
            a3 = tf.Print(a3, [a3], message='a3=')
            a4 = tf.Print(a4, [a4], message='a4=')
            a5 = tf.Print(a5, [a5, Phi], message='a5 and Phi=', summarize=10)
        return a1+a2+a3+a4+a5

    @params_as_tensors
    def _build_predict(self, Xnew, full_cov=False):
        M = tf.shape(self.X)[0]
        K = self.kern.K(self.X)
        Phi = tf.nn.softmax(self.logPhi)
        # try squashing Phi to avoid numerical errors
        Phi = (1 - 2e-6) * Phi + 1e-6
        sigma2 = self.likelihood.variance
        L = tf.cholesky(K) + tf.eye(M, dtype=settings.float_type) * settings.numerics.jitter_level
        W = tf.transpose(L) * tf.sqrt(tf.reduce_sum(Phi, 0)) / tf.sqrt(sigma2)
        P = tf.matmul(W, tf.transpose(W)) + tf.eye(M, dtype=settings.float_type)
        R = tf.cholesky(P)
        PhiY = tf.matmul(tf.transpose(Phi), self.Y)
        LPhiY = tf.matmul(tf.transpose(L), PhiY)
        c = tf.matrix_triangular_solve(R, LPhiY, lower=True) / sigma2
        Kus = self.kern.K(self.X, Xnew)
        tmp1 = tf.matrix_triangular_solve(L, Kus, lower=True)
        tmp2 = tf.matrix_triangular_solve(R, tmp1, lower=True)
        mean = tf.matmul(tf.transpose(tmp2), c)
        if full_cov:
            var = self.kern.K(Xnew) + tf.matmul(tf.transpose(tmp2), tmp2)\
                - tf.matmul(tf.transpose(tmp1), tmp1)
            shape = tf.stack([1, 1, tf.shape(self.Y)[1]])
            var = tf.tile(tf.expand_dims(var, 2), shape)
        else:
            var = self.kern.Kdiag(Xnew) + tf.reduce_sum(tf.square(tmp2), 0)\
                - tf.reduce_sum(tf.square(tmp1), 0)
            shape = tf.stack([1, tf.shape(self.Y)[1]])
            var = tf.tile(tf.expand_dims(var, 1), shape)
        return mean, var

    def build_KL(self, Phi):
        Bv_s = tf.squeeze(self.kern.kernels[0].Bv, squeeze_dims=[1])
        # pZ = pZ_construction_singleBP.make_matrix(self.t, Bv_s, self.phiPrior)
        return tf.reduce_sum(Phi * tf.log(Phi)) - tf.reduce_sum(Phi * tf.log(self.pZ))



import numpy as np
from matplotlib import pyplot as plt

def CalculateBranchingEvidence(d, Bsearch=None):
    """
    :param d: output dictionary from FitModel
    :param Bsearch: candidate list of branching points
    :return: posterior probability of branching at each point and log Bayes factor
    of branching vs not branching
    """
    if Bsearch is None:
        Bsearch = list(np.linspace(0.05, 0.95, 5)) + [1.1]
    # Calculate probability of branching at each point
    o = d['loglik'][:-1]
    pn = np.exp(o - np.max(o))
    p = pn/pn.sum()  # normalize

    # Calculate log likelihood ratio by averaging out
    o = d['loglik']
    Nb = o.size - 1
    if Nb != len(Bsearch) - 1:
        raise NameError('Passed in wrong length of Bsearch is %g- should be %g' % (len(Bsearch), Nb))
    obj = o[:-1]
    illmax = np.argmax(obj)
    llmax = obj[illmax]
    lratiostable = llmax + np.log(1 + np.exp(obj[np.arange(obj.size) != illmax]-llmax).sum()) - o[-1] - np.log(Nb)

    return {'posteriorBranching': p, 'logBayesFactor': lratiostable}


def PlotBGPFit(GPy, GPt, Bsearch, d, figsize=(5, 5), height_ratios= [5, 1], colorarray=['darkolivegreen', 'peru', 'mediumvioletred']):
    """
    Plot BGP model
    :param GPt: pseudotime
    :param GPy: gene expression. Should be 0 mean for best performance.
    :param Bsearch: list of candidate branching points
    :param d: output dictionary from FitModel
    :param figsize: figure size
    :param height_ratios: ratio of assignment plot vs posterior branching time plot
    :param colorarray: colors for each branch
    :return: dictionary of log likelihood, GPflow model, Phi matrix, predictive set of points,
    mean and variance, hyperparameter values, posterior on branching time
    """
    fig, axa = plt.subplots(2, 1, figsize=figsize, sharex=True,  gridspec_kw = {'height_ratios': height_ratios})
    ax = axa[0]
    y, pt, mul, ttestl = GPy, GPt, d['prediction']['mu'], d['prediction']['xtest']
    lw = 4
    for f in range(3):
        mu = mul[f]
        ttest = ttestl[f]
        col = colorarray[f]  # mean.get_color()
        ax.plot(ttest, mu, linewidth=lw, color=col, alpha=0.7)
        gp_num = 1  # can be 0,1,2 - Plot against this
        PhiColor = ax.scatter(pt, y, c=d['Phi'][:, gp_num], vmin=0., vmax=1, s=40, alpha=0.7)
    _ = fig.colorbar(PhiColor, ax=ax, orientation="horizontal")
    ax = axa[1]
    p = CalculateBranchingEvidence(d, Bsearch)['posteriorBranching']
    ax.stem(Bsearch[:-1], p)
    return fig, axa


def plotBranchModel(B, pt, Y, ttestl, mul, varl, Phi, figsizeIn=(5, 5), lw=3., fs=10, labels=None,
                    fPlotPhi=True, fPlotVar=False, ax=None, fColorBar=True, colorarray = ['darkolivegreen', 'peru', 'mediumvioletred']):
    ''' Plotting code that does not require access to the model but takes as input predictions. '''
    if(ax is None):
        fig = plt.figure(figsize=figsizeIn)
        ax = fig.gca()
    else:
        fig = plt.gcf()
    d = 0  # constraint code to be 1D for now
    for f in range(3):
        mu = mul[f]
        var = varl[f]
        ttest = ttestl[f]
        col = colorarray[f]  # mean.get_color()
        mean, = ax.plot(ttest, mu[:, d], linewidth=lw, color=col)
        if(fPlotVar):
            ax.plot(ttest.flatten(), mu[:, d] + 2 * np.sqrt(var.flatten()), '--', color=col, linewidth=lw)
            ax.plot(ttest, mu[:, d] - 2 * np.sqrt(var.flatten()), '--', color=col, linewidth=lw)
    v = ax.axis()
    ax.plot([B, B], v[-2:], '-m', linewidth=lw)
    # Plot Phi or labels
    if(fPlotPhi):
        gp_num = 1  # can be 0,1,2 - Plot against this
        PhiColor = ax.scatter(pt, Y[:, d], c=Phi[:, gp_num], vmin=0., vmax=1, s=40)
        if(fColorBar):
            fig.colorbar(PhiColor, label='GP {} assignment probability'.format(gp_num))
        else:
            return fig, PhiColor
    return fig


def predictBranchingModel(m, full_cov=False):
    ''' return prediction of branching model '''
    pt = m.t
    B = m.kern.kernels[0].Bv.value.flatten()
    l = np.min(pt)
    u = np.max(pt)
    mul = list()
    varl = list()
    ttestl = list()
    for f in range(1, 4):
        if(f == 1):
            ttest = np.linspace(l, B, 100)#[:, None]  # root
        else:
            ttest = np.linspace(B, u, 100)#[:, None]
        Xtest = np.hstack((ttest, ttest * 0 + f))
        # print('*************************************************************************')
        # print('Inside predictBranchingModel ...')
        if full_cov:
            mu, var = m.predict_f_full_cov(Xtest)
        else:
            mu, var = m.predict_f(Xtest)
        # print('mu', mu)
        idx = np.isnan(mu)
        # print('munan', mu[idx], var[idx], ttest[idx])
        assert np.all(np.isfinite(mu)), 'All elements should be finite but are ' + str(mu)
        assert np.all(np.isfinite(var)), 'All elements should be finite but are ' + str(var)
        mul.append(mu)
        varl.append(var)
        ttestl.append(ttest)
    return ttestl, mul, varl


def GetFunctionIndexListGeneral(Xin):
    ''' Function to return index list  and input array X repeated as many time as each possible function '''
    # limited to one dimensional X for now!
    assert Xin.shape[0] == np.size(Xin)
    indicesBranch = []

    XSample = np.zeros((Xin.shape[0], 2), dtype=float)
    Xnew = []
    inew = 0
    functionList = list(range(1, 4))  # can be assigned to any of root or subbranches, one based counting

    for ix, x in enumerate(Xin):
        XSample[ix, 0] = Xin[ix]
        XSample[ix, 1] = np.random.choice(functionList)
        # print str(ix) + ' ' + str(x) + ' f=' + str(functionList) + ' ' +  str(XSample[ix,1])
        idx = []
        for f in functionList:
            Xnew.append([x, f])  # 1 based function list - does kernel care?
            idx.append(inew)
            inew = inew + 1
        indicesBranch.append(idx)
    Xnewa = np.array(Xnew)
    return (Xnewa, indicesBranch, XSample)


def SetXExpandedBranchingPoint(XExpanded, B):
    ''' Return XExpanded by removing unavailable branches '''
    # before branching pt, only function 1
    X1 = XExpanded[np.logical_and(XExpanded[:, 0] <= B, XExpanded[:, 1] == 1).flatten(), :]
    # after branching pt, only functions 2 and 2
    X23 = XExpanded[np.logical_and(XExpanded[:, 0] > B, XExpanded[:, 1] != 1).flatten(), :]
    return np.vstack([X1, X23])



# coding: utf-8
import gpflow
import numpy as np
import tensorflow as tf
from . import assigngp_dense
from gpflow import settings
from gpflow.params import DataHolder
from gpflow.decors import params_as_tensors, autoflow


class AssignGPSparse(assigngp_dense.AssignGP):
    """
    Gaussian Process sparse regression, but where the index to which the data are
    assigned is unknown.

    let f be a vector of GP points (usually longer than the number of data)

        f ~ MVN(0, K)

    and let Z be an (unknown) binary matrix with a single 1 in each row. The
    likelihood is

       y ~ MVN( Z f, \sigma^2 I)

    That is, each element of y is a noisy realization of one (unknown) element
    of f. We use variational Bayes to infer the labels using a sparse prior
    over the Z matrix (i.e. we have narrowed down the choice of which function
    values each y is drawn from).

    """

    def __init__(self, t, XExpanded, Y, kern, indices, b, ZExpanded, fDebug=False, phiInitial=None, phiPrior=None):
        assigngp_dense.AssignGP.__init__(self, t, XExpanded, Y, kern, indices, b, fDebug=fDebug,
                                         phiInitial=phiInitial, phiPrior=phiPrior)
        # Do not treat inducing points as parameters because they should always be fixed.
        self.ZExpanded = DataHolder(ZExpanded)  # inducing points for sparse GP. Same as XExpanded
        assert ZExpanded.shape[1] == XExpanded.shape[1]

    @params_as_tensors
    def _build_likelihood(self):
        if self.fDebug:
            print('assignegp_denseSparse compiling model (build_likelihood)')
        N = tf.cast(tf.shape(self.Y)[0], dtype=settings.float_type)
        M = tf.shape(self.ZExpanded)[0]
        D = tf.cast(tf.shape(self.Y)[1], dtype=settings.float_type)

        Phi = tf.nn.softmax(self.logPhi)
        # try squashing Phi to avoid numerical errors
        Phi = (1-2e-6) * Phi + 1e-6

        sigma2 = self.likelihood.variance
        sigma = tf.sqrt(self.likelihood.variance)
        Kuu = self.kern.K(self.ZExpanded) + tf.eye(M, dtype=settings.float_type) * settings.numerics.jitter_level
        Kuf = self.kern.K(self.ZExpanded, self.X)

        Kdiag = self.kern.Kdiag(self.X)
        L = tf.cholesky(Kuu)
        A = tf.reduce_sum(Phi, 0)
        LiKuf = tf.matrix_triangular_solve(L, Kuf)
        W = LiKuf * tf.sqrt(A) / sigma
        P = tf.matmul(W, tf.transpose(W)) + tf.eye(M, dtype=settings.float_type)
        traceTerm = -0.5 * tf.reduce_sum(Kdiag * A) / sigma2 + 0.5 * tf.reduce_sum(tf.square(W))
        R = tf.cholesky(P)
        tmp = tf.matmul(LiKuf, tf.matmul(tf.transpose(Phi), self.Y))
        c = tf.matrix_triangular_solve(R, tmp, lower=True) / sigma2
        if(self.fDebug):
            # trace term should be 0 for Z=X (full data)
            traceTerm = tf.Print(traceTerm, [traceTerm], message='traceTerm=', name='traceTerm', summarize=10)

        self.bound = traceTerm - 0.5*N*D*tf.log(2 * np.pi * sigma2)\
            - 0.5*D*tf.reduce_sum(tf.log(tf.square(tf.diag_part(R))))\
            - 0.5*tf.reduce_sum(tf.square(self.Y)) / sigma2\
            + 0.5*tf.reduce_sum(tf.square(c))\
            - self.build_KL(Phi)

        return self.bound

    @params_as_tensors
    def _build_predict(self, Xnew, full_cov=False):
        M = tf.shape(self.ZExpanded)[0]

        Phi = tf.nn.softmax(self.logPhi)
        # try squashing Phi to avoid numerical errors
        Phi = (1-2e-6) * Phi + 1e-6

        sigma2 = self.likelihood.variance
        sigma = tf.sqrt(sigma2)
        Kuu = self.kern.K(self.ZExpanded) + tf.eye(M, dtype=settings.float_type) * settings.numerics.jitter_level
        Kuf = self.kern.K(self.ZExpanded, self.X)
        L = tf.cholesky(Kuu)

        p = tf.reduce_sum(Phi, 0)
        LiKuf = tf.matrix_triangular_solve(L, Kuf)
        W = LiKuf * tf.sqrt(p) / sigma
        P = tf.matmul(W, tf.transpose(W)) + tf.eye(M, dtype=settings.float_type)
        R = tf.cholesky(P)
        tmp = tf.matmul(LiKuf, tf.matmul(tf.transpose(Phi), self.Y))
        c = tf.matrix_triangular_solve(R, tmp, lower=True) / sigma2

        Kus = self.kern.K(self.ZExpanded, Xnew)
        tmp1 = tf.matrix_triangular_solve(L, Kus, lower=True)
        tmp2 = tf.matrix_triangular_solve(R, tmp1, lower=True)
        mean = tf.matmul(tf.transpose(tmp2), c)
        if full_cov:
            var = self.kern.K(Xnew) + tf.matmul(tf.transpose(tmp2), tmp2)\
                - tf.matmul(tf.transpose(tmp1), tmp1)
            shape = tf.stack([1, 1, tf.shape(self.Y)[1]])
            var = tf.tile(tf.expand_dims(var, 2), shape)
        else:
            var = self.kern.Kdiag(Xnew) + tf.reduce_sum(tf.square(tmp2), 0)\
                - tf.reduce_sum(tf.square(tmp1), 0)
            shape = tf.stack([1, tf.shape(self.Y)[1]])
            var = tf.tile(tf.expand_dims(var, 1), shape)
        return mean, var
